/*
 * Copyright (c) 2003-2005 Erez Zadok
 * Copyright (c) 2003-2005 Charles P. Wright
 * Copyright (c) 2003-2005 Mohammad Nayyer Zubair
 * Copyright (c) 2003-2005 Puja Gupta
 * Copyright (c) 2003-2005 Harikesavan Krishnan
 * Copyright (c) 2003-2005 Stony Brook University
 * Copyright (c) 2003-2005 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */
/*
 *  $Id: rdstate.c,v 1.3 2005/01/09 21:43:25 ezk Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include "fist.h"
#include "unionfs.h"

/* This file contains the routines for maintaining readdir state. */

/* This is a tuning parameter that tells us roughly how big to make the
 * hash table in directory entries per page.  This isn't perfect, but
 * at least we get a hash table size that shouldn't be too overloaded.
 * The following averages are based on my home directory.
 * 14.44693	Overall
 * 12.29 	Single Page Directories
 * 117.93	Multi-page directories
 */
#define DENTPAGE 4096
#define DENTPERONEPAGE 12
#define DENTPERPAGE 118
#define MINHASHSIZE 1
int guesstimate_hash_size(struct inode *inode) {
	struct inode *hidden_inode;
	int bindex;
	int hashsize = MINHASHSIZE;

	if (itopd(inode)->uii_hashsize > 0) {
		return itopd(inode)->uii_hashsize;
	}

	for (bindex = ibstart(inode); bindex <= ibend(inode); bindex++) {
		if ((hidden_inode = itohi_index(inode, bindex))) {
			if (hidden_inode->i_size == DENTPAGE) {
				hashsize += DENTPERONEPAGE;
			} else {
				hashsize += (hidden_inode->i_size / DENTPAGE) * DENTPERPAGE;
			}
		}
	}

	return hashsize;
}

int init_rdstate(struct file *file) {
	int i = 0;
	int hashsize;
	int mallocsize = sizeof(struct unionfs_dir_state);

	PASSERT(file);
	ASSERT(ftopd(file)->rdstate == NULL);
	PASSERT(file->f_dentry);
	PASSERT(file->f_dentry->d_inode);

	hashsize = guesstimate_hash_size(file->f_dentry->d_inode);
	mallocsize += hashsize * sizeof(struct list_head);
	/* Round it up to the next highest power of two. */
	mallocsize--;
	mallocsize |= mallocsize >> 1;
	mallocsize |= mallocsize >> 2;
	mallocsize |= mallocsize >> 4;
	mallocsize |= mallocsize >> 8;
	mallocsize |= mallocsize >> 16;
	mallocsize++;
	if (mallocsize > PAGE_SIZE) { /* This should give us thousands of entries anyway. */
		mallocsize = PAGE_SIZE;
	}
	hashsize = (mallocsize - sizeof(struct unionfs_dir_state)) / sizeof(struct list_head);

	ftopd(file)->rdstate = KMALLOC(mallocsize, GFP_KERNEL);
	if (!ftopd(file)->rdstate) {
		return -ENOMEM;
	}

	ftopd(file)->rdstate->uds_key = 0;
	ftopd(file)->rdstate->uds_offset = 1;
	ftopd(file)->rdstate->uds_access = jiffies;
	ftopd(file)->rdstate->uds_bindex = fbstart(file);
	ftopd(file)->rdstate->uds_dirpos = 0;
	ftopd(file)->rdstate->uds_hashentries = 0;
	ftopd(file)->rdstate->uds_size = hashsize;
	for (i = 0; i < ftopd(file)->rdstate->uds_size; i++)
		INIT_LIST_HEAD(&(ftopd(file)->rdstate->uds_list[i]));

	return 0;
}

void free_rdstate(struct unionfs_dir_state *state) {
    struct filldir_node *tmp;
    int i;

    PASSERT(state);

    for (i = 0; i < state->uds_size; i++) {
	struct list_head *head = &(state->uds_list[i]);
	struct list_head *pos, *n;

	/* traverse the list and deallocate space */
	list_for_each_safe(pos, n, head) {
	    tmp = list_entry(pos, struct filldir_node, file_list);
	    list_del(&tmp->file_list);
	    free_filldir_node(tmp);
	}
    }

    KFREE(state);
}

/* This is a kmem_cache_t for filldir nodes, because we allocate a lot of them
 * and they shouldn't waste memory.  If the node has a small name (as defined
 * by the dentry structure), then we use an inline name to preserve kmalloc
 * space. */
static kmem_cache_t *unionfs_filldir_cachep;
int init_filldir_cache() {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
	unionfs_filldir_cachep = kmem_cache_create("unionfs_filldir", sizeof(struct filldir_node), 0, SLAB_RECLAIM_ACCOUNT, NULL, NULL);
#else
	unionfs_filldir_cachep = kmem_cache_create("unionfs_filldir", sizeof(struct filldir_node), 0, 0, NULL, NULL);
#endif
	if (!unionfs_filldir_cachep) {
		return -ENOMEM;
	}
	return 0;
}
void destroy_filldir_cache() {
	if (kmem_cache_destroy(unionfs_filldir_cachep)) {
		printk(KERN_INFO "unionfs_filldir_cache: not all structures were freed\n");
	}
	return;
}

inline struct filldir_node *alloc_filldir_node(int namelen, int hash, int bindex, const char *name) {
	struct filldir_node *newnode;

	newnode = (struct filldir_node *)kmem_cache_alloc(unionfs_filldir_cachep, SLAB_KERNEL);
        if (!newnode) {
		goto out;
        }

	INIT_LIST_HEAD(&newnode->file_list);
	newnode->namelen = namelen;
	newnode->hash = hash;
	newnode->bindex = bindex;

	if (namelen < DNAME_INLINE_LEN_MIN) {
	    newnode->name = newnode->iname;
	} else {
	    newnode->name = (char *) KMALLOC(namelen + 1, GFP_KERNEL);
	    if (!newnode->name) {
		kmem_cache_free(unionfs_filldir_cachep, newnode);
		newnode = NULL;
		goto out;
	    }
	}

	memcpy(newnode->name, name, namelen);
	newnode->name[namelen] = '\0';

out:
	return newnode;
}

inline void free_filldir_node(struct filldir_node *node) {
	PASSERT(node);
	if (node->namelen >= DNAME_INLINE_LEN_MIN) {
		KFREE(node->name);
	}
	kmem_cache_free(unionfs_filldir_cachep, node);
}
